{****************************************************************************
*            } program CodCon; { (c) 2000 Michal Cihar                     *
 ****************************************************************************
This is demonstation of usage CodeConvTools unit

Useful tool for converting files between many code-pages.
You must specify two file names for conversion - first is allways used for
input and second fou output, more filenames are ignored. Output file can be
ommited only if there is also parameter /R which means that input file is
overwritten by file with new code-page. You can specify input and output
code-page. If no input code-page is specified so autodetection is performed.
No output code-page means that file will be converted to pure Ascii.
Note that autodetection of input code-page can't be absolutely exact
especially some code-pages are very difficult to detect (Dragon).
Currently supported code-pages are listed in CodeConvTools.pas
}
//FPC compatibility switches
{$IFDEF FPC}
{$MODE DELPHI}
{$ASMMODE intel}
{$ENDIF}

uses
  SysUtils,
  CodeConvTools,
  Windows,
  CommonData in '..\CommonData.pas';
{$R *.res}
{$APPTYPE CONSOLE}

const
  CPFrom: TCodePage = cp1250;
  CPTo: TCodePage = cpAscii;
  ToCase: TCase = CaseNoChange;
  ToEOL: TEOL = EOLMakeAsCP;
  DetectCP: Boolean = True;
  WriteOver: Boolean = False;

var
  FileIn, FileOut, InName, OutName: string;
  ConvertThread: TConvertThread;
  I, F: Integer;
  Param: string;

procedure Help;
begin
  WriteLn('This program must be spread free, any modifications are allowed, while' + #13#10 +
    'modifications are notified to me (e-mail: ' + AuthorEmail + ') and my copyright' + #13#10 +
    'stays on same visible place.' + #13#10);

  WriteLn('Code page converter can be used in this style:');
  WriteLn(' CodeConv FileIn FileOut options');
  WriteLn('  options: FileOut code page: /K ... ', CodePageNames[cp895]);
  WriteLn('                              /W ... ', CodePageNames[cp1250]);
  WriteLn('                              /L ... ', CodePageNames[cp852]);
  WriteLn('                              /I ... ', CodePageNames[cpISO88592]);
  WriteLn('                              /1 ... ', CodePageNames[cpISO88591]);
  WriteLn('                              /8 ... ', CodePageNames[cpKOI8]);
  WriteLn('                              /C ... ', CodePageNames[cpCork]);
  WriteLn('                              /M ... ', CodePageNames[cpMac]);
  WriteLn('                              /D ... ', CodePageNames[cpDOS]);
  WriteLn('                              /R ... ', CodePageNames[cpDragon]);
  WriteLn('                              /A ... ', CodePageNames[cpAscii]);
  WriteLn('                              /B ... ', CodePageNames[cpIBM2]);
  WriteLn('                              /0 ... ', CodePageNames[cp850]);
  WriteLn('           FileIn code page: by default is performed autodetection, but can be');
  WriteLn('                             overridden by /Fx where x is char for code page');
  WriteLn('                             same as for FileOut');
  Write('Press enter for more options...');
  ReadLn;
  WriteLn('           other formatting: /S or /CL or /C1 ... convert all to lowercase');
  WriteLn('                             /U or /CH or /C2 ... convert all to uppercase');
  WriteLn('                             /CN or /C0 ... don''t change case - DEFAULT');
  WriteLn;
  WriteLn('                             /Ex ... change end of line format of FileOut');
  WriteLn('                               x can be:');
  WriteLn('                                 N, 0 - don''t modify (same as in FileIn)');
  WriteLn('                                 A, 1 - change according to codepage (DEFAULT)');
  WriteLn('                                 C, 2 - CR (Macintosh)');
  WriteLn('                                 L, 3 - LF (UNIX, Linux)');
  WriteLn('                                 B, 4 - CRLF (DOS, Windows)');
  WriteLn('           other switches: /O ... only modifies FileIn (no FileOut needed)');
  WriteLn('Other notes:  ');
  WriteLn('  Parameters can be in any order, first filename is allways used for input.');
  WriteLn('You must specify two filenames or one filename and /O switch. "-" and "/"');
  WriteLn('are treated as same.');
  WriteLn('  Latest version of this program and another my programs can be found on');
  WriteLn(AuthorWeb + ', any comments or bugreports send to ' + AuthorEmail + '. You');
  WriteLn('can also send me information about other code pages that I should include');
  WriteLn('in this program.');
  WriteLn('  If you thing console version is not good for you, note that also "clicking"');
  WriteLn('windows version exists on ' + AuthorWeb);
  WriteLn('Press enter to exit...');
  ReadLn;
  Halt(255);
end;



begin
  CreateMutex(nil, False, MutexName);

  WriteLn('CICOM ' + AppName + ' (32-bit) version ' + VersionString + ' copyright (c) 2000 Michal Cihar, freeware' + #13#10);

  F := 0;
  if ParamCount = 0 then Help;
  for I := 1 to ParamCount do
  begin
    Param := ParamStr(I);
    if Param = '--help' then
      Help
    else
      if Param[1] in ['/', '-'] then
      begin
        if Length(Param) = 2 then
          case UpCase(Param[2]) of
            '?', 'H': Help;

            'K': CPTo := cp895;
            'W': CPTo := cp1250;
            'A': CPTo := cpAscii;
            'D': CPTo := cpDOS;
            'M': CPTo := cpMac;
            'I': CPTo := cpISO88592;
            '1': CPTo := cpISO88591;
            'L': CPTo := cp852;
            'B': CPTo := cpIBM2;
            '0': CPTo := cp850;
            '8': CPTo := cpKOI8;
            'C': CPTo := cpCork;
            'R': CPTo := cpDragon;

            'S': ToCase := CaseLow;
            'U': ToCase := CaseHigh;
            'O': WriteOver := True;
          else
            Help;
          end
        else
          if Length(Param) = 3 then
          begin
            if UpCase(Param[2]) = 'F' then
            begin
              DetectCP := False;
              case UpCase(Param[3]) of
                'K': CPFrom := cp895;
                'W': CPFrom := cp1250;
                'B': CPFrom := cpIBM2;
                '0': CPFrom := cp850;
                'L': CPFrom := cp852;
                'I': CPFrom := cpISO88592;
                '1': CPFrom := cpISO88591;
                '8': CPFrom := cpKOI8;
                'C': CPFrom := cpCork;
                'D': CPFrom := cpDOS;
                'R': CPFrom := cpDragon;
                'M': CPFrom := cpMac;
                'A': CPFrom := cpAscii;
              else
                Help;
              end;
            end
            else
              if UpCase(Param[2]) = 'E' then
              begin
                case UpCase(Param[3]) of
                  '0', 'N': ToEOL := EOLNoChange;
                  '1', 'A': ToEOL := EOLMakeAsCP;
                  '2', 'C': ToEOL := EOLMakeCR;
                  '3', 'L': ToEOL := EOLMakeLF;
                  '4', 'B': ToEOL := EOLMakeCRLF;
                else
                  Help;
                end;
              end
              else
                if UpCase(Param[2]) = 'C' then
                begin
                  case UpCase(Param[3]) of
                    '0', 'N': ToCase := CaseNoChange;
                    '1', 'L': ToCase := CaseLow;
                    '2', 'H': ToCase := CaseHigh;
                  else
                    Help;
                  end;
                end
                else
                  Help;
          end
          else
            Help;
      end
      else
      begin
        if F = 0 then
        begin
          if not (FileExists(Param)) then
          begin
            WriteLn('ERROR: File ' + Param + ' not found');
            WriteLn;
            Halt(1);
          end
          else
          begin
            FileIn := ExpandFileName(Param);
            InName := ExtractFileName(FileIn);
            Inc(F);
          end;
        end
        else
          if F = 1 then
            if UpperCase(FileIn) = UpperCase(ExpandFileName(Param)) then
              WriteOver := True
            else
            begin
              FileOut := ExpandFileName(Param);
              OutName := ExtractFileName(FileOut);
              Inc(F);
            end;
      end;
  end;
  if WriteOver then
    FileOut := FileIn
  else
    if (F = 1) and not (WriteOver) then Help;

  if WriteOver then
    WriteLn('Converting file ', InName, ' to itself')
  else
    WriteLn('Converting file ', InName, #13#10'to ', OutName);
  if DetectCP then
  begin
    Write('Detecting code page of source...'#13);
    CPFrom := DetectCodepage(FileIn);
    WriteLn('For file ', InName, ' was detected code page: ', CodePageNames[CPFrom]);
  end
  else
    WriteLn('For file ', InName, ' using code page: ', CodePageNames[CPFrom]);

  ConvertThread := TConvertThread.Create(nil, nil, FileIn, FileOut, CPFrom, CPTo, ToCase, ToEOL);
  ConvertThread.Resume;

  Write('Converting to code page: ', CodePageNames[CPTo]);
  case ToEOL of
    EOLMakeAsCP:
      case CRLFs[CPTo] of
        EOLMakeCR: Write(' [CR]');
        EOLMakeLF: Write(' [LF]');
        EOLMakeCRLF: Write(' [CRLF]');
      end;
    EOLMakeCR: Write(' [CR]');
    EOLMakeLF: Write(' [LF]');
    EOLMakeCRLF: Write(' [CRLF]');
  end;
  case ToCase of
    CaseNoChange: WriteLn;
    CaseLow: WriteLn(' with conversion to lower case');
    CaseHigh: WriteLn(' with conversion to upper case');
  end;


  Write('0   % converted'#13);
  while ((not (ConvertThread.WasTerminated)) and ((ConvertThread.InputFileSize = 0))) do
    Sleep(10);
  while not (ConvertThread.WasTerminated) do
  begin
    Write(Round((ConvertThread.CurrentPosition / ConvertThread.InputFileSize) * 100), #13);
    Sleep(750); //greater value means a bit faster conversion and less fluent display
  end;
  WriteLn('Conversion succesfully completed. Converted ', ConvertThread.CurrentPosition , ' bytes.' + #13#10 +
    'Thanks for using my program. More of them can be found on ' + AuthorWeb + #13#10);
  ConvertThread.Free;
end.

