{********************************* i n f o **********************************}
{*                         } unit CodeConvTools; {                          *}
{*                          (c) 2000 Michal Cihar                           *}
{******************************** e m a i l *********************************}
{*                            <cihar@email.cz>                              *}
{***************************** i n t e r n e t ******************************}
{*                          http://cicom.tsx.org                            *}
{********************************** i c q ***********************************}
{*                              UIN: 33471395                               *}
{************************** d e s c r i p t i o n ***************************}
{*    This unit enables you to use czech text files in (almost) any code    *}
{*  page (encoding). Now 12 output code pages are supported, so I hope      *}
{*  there is no other code page used. If there is any please contact me     *}
{*  and I will add this. I hope, This unit will support ALL (thousands :o)) *}
{*  Czech code pages.                                                       *}
{*    You don't have to know which code page has opened text file, because  *}
{*  this unit cpntain code page detection algorithm. Although this          *}
{*  detection works with most of files correctly, but you shouldn't         *}
{*  absolutely believe to this and give a chance to user to correct bad     *}
{*  detection.                                                              *}
{******************************* m a n u a l ********************************}
{* TYPES                                                                    *}
{*   TCodePage                                                              *}
{*      Ordinal type with one value for each encoding:                      *}
{*                  Kamenicky, Windows, PCLatin2, CP852, ISO8859,           *}
{*                  KOI8, Cork, Dragon, Mac, Ascii                          *}
{*                                                                          *}
{*   TCase                                                                  *}
{*      Ordinal type for case changing values                               *}
{*                  CaseNoChange, CaseLow, CaseHigh                         *}
{*                                                                          *}
{*   TEOL                                                                   *}
{*      Ordinal type for EOL characters conversion                          *}
{*                  EOLNoChange, EOLMakeAsCP, EOLMakeCR, EOLMakeLF,         *}
{*                  EOLMakeCRLF                                             *}
{*                                                                          *}
{* CONSTANTS                                                                *}
{*  read-only                                                               *}
{*   FirstCodePage, LastCodePage                                            *}
{*      First and last code page in tCodePage. Last should be always Ascii. *}
{*   CodePageCount                                                          *}
{*      Count of code pages in tCodePage                                    *}
{*   SingleBorders, DoubleBorders, ThickBorders                             *}
{*      Characters used for borders in DOS text files                       *}
{*   MaxChars                                                               *}
{*      Number of chars in conversion table                                 *}
{*  read/write                                                              *}
{*   ConversionTable                                                        *}
{*      Table used for conversions, change advisedly                        *}
{*   CodePageDescriptionEN, CodePageDescriptionCZ                           *}
{*      Code pages description in english and czech                         *}
{*   CodePageNames                                                          *}
{*      Short (limited by 25 chars) code page names                         *}
{*   DetectRead                                                             *}
{*      How many chars use to detect code page                              *}
{*   IgnoreBorders                                                          *}
{*      Ignore borders duing detection, best results are mostly when this   *}
{*      is true.                                                            *}
{*                                                                          *}
{*                                                                          *}
{*                                                                          *}
{****************************** l i c e n c e *******************************}
{*    This source is freeware with some small restrictions. For details     *}
{*  look into file LICENSE.TXT, which you should get with this source.      *}
{*************************** c o d e   p a g e s ****************************}
{*    Currently supported code pages:                                       *}
{*                                                                          *}
{*     Kamenicky (895)                                                      *}
{*     Windows (1250)                                                       *}
{*     DOS Latin 2 (852)                                                    *}
{*     ISO-8859-2 (Unix/Linux)                                              *}
{*     ISO-8859-1 (Unix/Linux)                                              *}
{*     KOI8-CS (T602)                                                       *}
{*     Cork (TeX)                                                           *}
{*     Dragon                                                               *}
{*     Macintosh                                                            *}
{*     DOS (437) - all chars that are in standard DOS code page             *}
{*     Other Latin 2 - I thing that this code is used nowhere, but          *}
{*                     somewhere I found this and I want this unit supports *}
{*                     all czech code pages ever used :o)                   *}
{*     Ascii (7 bit)                                                        *}
{*                                                                          *}
{*    If you thing that here is some code page missing, please feel free    *}
{*    to contact me via email <cihar@email.cz> about this.                  *}
{*                                                                          *}
{****************************** h i s t o r y *******************************}
{*  version : date                                                          *}
{*            changes (+ added, - removed, * changed)                       *}
{*  1.5 : 1.1.2001                                                          *}
{*        * improved autodetection                                          *}
{*        *                                                                 *}
{*        *                                                                 *}
{*        *                                                                 *}
{*  1.2 : 19.10.2000                                                        *}
{*        * ConvertFile is done using TConvertThread                        *}
{*        * changed style of buffered reading and writing                   *}
{*        + added code page PC Latin 1                                      *}
{*        + added conversion of some special chars (quotes, dashes,         *}
{*          non breaking space)                                             *}
{*        * codepage Other Latin 2 renamed to IBM2 (IBM Latin 2)            *}
{*        * speeded up conversion                                           *}
{*  1.2 : 18.9.2000                                                         *}
{*        * corrected table for ISO-8859-1 encoding                         *}
{*  1.1 : 14.9.2000                                                         *}
{*        * big speed improvements for file conversion (more than 30 times  *}
{*          faster), because using buffer also for output file              *}
{*        + now can be compiled with Free Pascal Compiler                   *}
{*        + borders and specials conversion                                 *}
{*        + thraeded conversion                                             *}
{*        + DOS (437) code page                                             *}
{*        + ISO 8859-1 code page                                            *}
{*        + cpAny code page                                                 *}
{*        * fixed few bugs in tables                                        *}
{*        * some code optimalisations                                       *}
{*        + comments and code made cleaner                                  *}
{*        - UpCase function (useles - same defined in pascal)               *}
{*        * some declarations moved to implemenation section                *}
{*  1.0 : 24.3.2000                                                         *}
{*        + documentation                                                   *}
{*        * corrected names of code pages                                   *}
{*        + functions DetectCodePage, ConvertChar, ConvertString            *}
{*        + arrays CodePageDescriptionEN, CodePageDescriptionCZ             *}
{*        + array CodePageNames                                             *}
{*        * a bit reorganised structures                                    *}
{*  This unit is based on my DOS program CodeConv 1.0                       *}
{****************************************************************************}
// FPC compatibility modes:
{$IFDEF FPC}
{$SMARTLINK ON}
{$MODE DELPHI}
{$ASMMODE intel}
{$ENDIF}
interface
uses Classes;
type

  TCodePage = (
   // Supported code page types
   // they are described in arrays: CodePageNames CodePageDescriptionXX (XX is language)
    cp895,
    cp1250,
    cp852,
    cpISO88592,
    cpISO88591,
    cpKOI8,
    cpCork,
    cpDragon,
    cpMac,
    cpDOS,
    cpIBM2,
    cp850,
    cpAscii,
    cpAny);
       // This code page has special magic:
       //  - when used in normal conversion is treated as cpAscii
       //  - when used as source code page for file, autodetection is performed

  TCase = (
   // Case conversion type
    CaseNoChange,
    CaseLow,
    CaseHigh);

  TEOL = (
   // End of lin conversion type
    EOLNoChange,
       // Don't change
    EOLMakeAsCP,
       // Change according to code page
    EOLMakeCR,
       // Change to CR {#10)
    EOLMakeLF,
       // Change to LF (#13)
    EOLMakeCRLF);
       // Change to CRLF (#13#10)

  TConvertArray = array[Char] of Char;


  TBeforeTerminateProc = procedure of object;
  TProgresProc = procedure (const Done, Total:Cardinal) of object;
   // Object progress procedure
  TProgresProc_ = procedure (const Done, Total:Cardinal);
   // Non object progress procedure


  TConvertThread = class (TThread)
  private
  protected
    ThreadProgProc:TProgresProc;
    ThreadBeforeTerminateProc:TBeforeTerminateProc;
    procedure CallProgress;
    procedure CallBeforeTerminateProc;
  public
    FFileIn, FFileOut:string;
    FCp1:TCodePage;
    FCp2:TCodePage;
    FChangeCase:TCase;
    FChangeEOL:TEOL;

    ThreadProgressStep:Cardinal;

    CompareFiles:Boolean;
    OutName:string;

    CurrentPosition:Cardinal;
    InputFileSize:Cardinal;

    WasTerminated:Boolean;

    constructor Create (
      ProgProc:TProgresProc;
          // Progress procedure
      BeforeTerminateProc:TBeforeTerminateProc;
          // Progress procedure
      FileIn:string;
          // File to convert (source)
      FileOut:string;
          // Output file (can be sane as source, then temporary file is used while converting)
      var CP1:TCodePage;
          // Source code page
      const CP2:TCodePage;
          // Target code page
      const ChangeCase:TCase;
          // Whether to change case of letters
      ChangeEOL:TEOL);
          // Whether to change end of line characters
    procedure Execute; override;
      // Procedure used by thread, to start thread execution use ConvertThread.Resume;
  end;

function FastPos (P:Pointer; BufLen:Cardinal; Chr:Char):Cardinal; assembler;

procedure GenerateConvertArray (const CP1, CP2:TCodePage; const C:TCase; var CA:TConvertArray);

function LoCase (Ch:Char):Char;
// Lower case of char (only for a-z without accents)

function DetectCodepage (
// Detects code page of file (should be text)
  const FileName:string
    // File which code page we want to detect
  ):TCodePage;
    // Resulting code page

function ConvertChar (
// Converts char between code pages
  const Ch:Char;
    // Char to convert
  const CP1:TCodePage;
    // Source code page
  const CP2:TCodePage;
    // Target code page
  const ChangeCase:TCase
    // Whether to change case of char
  ):Char;
    // Converted char

function ConvertString (
// Converts string between code pages
  const S:string;
    // String to convert
  const CP1:TCodePage;
    // Source code page
  const CP2:TCodePage;
    // Target code page
  const ChangeCase:TCase
    // Whether to change case of string
  ):string;
    // Converted string

function ConvertFile (
// Converts file between code pages
  FileIn:string;
    // File to convert (source)
  FileOut:string;
    // Output file (can be sane as source, then temporary file is used while converting)
  var CP1:TCodePage;
    // Source code page
  const CP2:TCodePage;
    // Target code page
  const ChangeCase:TCase;
    // Whether to change case of letters
  ChangeEOL:TEOL
    // Whether to change end of line characters
  ):Cardinal;
    // Size of converted file


// Redefinable constants
const
   //Name of temporary file used while converting file to same name
  TemporaryFileName:string =
  '$CodeConv temporary file - you can delete this if CodeConv is not running$.tmp';

   // How often are called progress procedures
   // if 0 then when initializing conversion, ProgressStep is set to FileSize div 100
  ProgressStep:Cardinal = 0;

   // How many chars use to detect code page
  DetectRead:Cardinal = 10000;
   // Ignore borders duing detection
  IgnoreBorders:Boolean = True;

   // Switches for conversion
  ConvertBorders:Boolean = True;
  ConvertSpecials:Boolean = True;

  ProgresProc:TProgresProc = nil;
  ProgresProc_:TProgresProc_ = nil;

const
// Informational constants
   // First and last code page for standard use - without cpAny
  FirstCodePage = cp895;
  LastCodePage = cpAscii;
  CodePageCount = Ord (High (TCodePage)) - 1;

   // Short code page names
  CodePageNames:array[TCodePage] of string[23] = (
    {cp895}'Kamenicky (895)',
    {cp1250}'Windows (1250)',
    {cp852}'DOS Latin 2 (852)',
    {cpISO88592}'ISO-8859-2 (Unix/Linux)',
    {cpISO88591}'ISO-8859-1 (Unix/Linux)',
    {cpKOI8}'KOI8-CS (T602)',
    {cpCork}'Cork (TeX T1)',
    {cpDragon}'Dragon',
    {cpMac}'Macintosh',
    {cpDOS}'DOS (437)',
    {cpIBM2}'IBM 852',
    {cp850}'DOS Latin 1 (850)',
    {cpAscii}'Ascii (7 bit)',
    {cpAny}'');
   // Long english code pages descriptions
  CodePageDescriptionEN:array[TCodePage] of string[50] = (
    {cp895}'De facto MS-DOS standard,
    {cp1250}'Microsoft Windows standard,
    {cp852}'Microsoft DOS standard for Latin 2',
    {cpISO88592}'Used in Unixes, Linux etc. Should be true standard,
    {cpISO88591}'Used in Unixes, Linux etc. for Latin 1',
    {cpKOI8}'Old standard for big hall computers, used in T602',
    {cpCork}'TeX standard from Cork',
    {cpDragon}'Only for program Dragon',
    {cpMac}'Standard in Macintosh computers',
    {cpDOS}'Standard ASCII 8-bit',
    {cpIBM2}'IBM charset for Latin 2 ??',
    {cp850}'Microsoft DOS standard for Latin 1',
    {cpAscii}'No accents',
    {cpAny}'Any code page');
   // Long czech code pages descriptions (cp1250)
  CodePageDescriptionCZ:array[TCodePage] of string[45] = (
    {cp895}'Nejpouvanj kdovn pro DOS',
    {cp1250}'Standard pro Microsoft Windows',
    {cp852}'Standard pro Microsoft DOS pro Latin 2',
    {cpISO88592}'Pouv Unix, Linux apod. Opravdov standard',
    {cpISO88591}'Pouv Unix, Linux apod. pro Latin 1',
    {cpKOI8}'Pro slov potae a T602',
    {cpCork}'Norma pro TeX z Corku',
    {cpDragon}'Pouito jen v programu Dragon',
    {cpMac}'Standard pro Macintoshe',
    {cpDOS}'Standardn ASCII 8-bit',
    {cpIBM2}'Kdovn IBM pro Latin 2 ??',
    {cp850}'Standard pro Microsoft DOS pro Latin 1',
    {cpAscii}'dn rky ani hky',
    {cpAny}'Jakkoliv');
   // Case names
  CaseName:array[TCase] of string[12] = (
    'Don''t change',
    'Low case',
    'High case');
   // EOL names
  EOLNames:array[TEOL] of string[17] = (
    'Don''t change',
    'Make as code page',
    'Make CR',
    'Make LF',
    'Make CRLF');

// Conversion tables
//  tables here are made from tables at the end of this file, but this tables enables faster conversion
   // Chars conversion table
  MaxChars = 44;
  cvtTable:array[TCodePage] of array[1..MaxChars] of Char = (
{Order:
   žؚݞ

   Missing chars: 
   }
   {TODO add }


    {cp895} #160#143#132#142#135#128#131#133#130#144#136#137#161#139#141#138#140#156#164#165#162#149#147#167#148#153#170#171#169#158#168#155#159#134#163#151#150#166#129#154#152#157#145#146,
    {cp1250} #225#193#228#196#232#200#239#207#233#201#236#204#237#205#229#197#190#188#242#210#243#211#244#212#246#214#224#192#248#216#154#138#157#141#250#218#249#217#252#220#253#221#158#142,
    {cp852} #160#181#132#142#159#172#212#210#130#144#216#183#161#214#146#145#150#149#229#213#162#224#147#226#148#153#234#232#253#252#231#230#156#155#163#233#133#222#129#154#236#237#167#166,
    {cpISO88592} #225#193#228#196#232#200#239#207#233#201#236#204#237#205#229#197#181#165#242#210#243#211#244#212#246#214#224#192#248#216#185#169#187#171#250#218#249#217#252#220#253#221#190#174,
    {cpISO88591} #$E1#$C1#$E4#$C4#$63#$43#$64#$44#$E9#$C9#$65#$45#$ED#$CD#$6C#$4C#$6C#$4C#$6E#$4E#$F3#$D3#$F4#$D4#$F6#$D6#$72#$52#$72#$52#$73#$53#$74#$54#$FA#$DA#$75#$55#$FC#$DC#$FD#$DD#$7A#$5A,
    {cpKOI8} #193#225#209#241#195#227#196#228#215#247#197#229#201#233#203#235#204#236#206#238#207#239#208#240#205#237#198#230#210#242#211#243#212#244#213#245#202#234#200#232#217#249#218#250,
    {cpCork} #225#193#228#196#163#131#164#132#233#201#165#133#237#205#168#136#169#137#172#140#243#211#244#212#246#214#175#143#176#144#178#146#180#148#250#218#183#151#252#220#253#221#186#154,
    {cpDragon} #$C0#$C1#097#065#$C2#$C3#$C4#$C5#$C6#$C7#$C8#$C9#$CA#$CB#108#076#108#076#$CC#$CD#$CE#$CF#111#079#111#079#114#082#$D0#$D1#$D2#$D3#$D4#$D5#$D6#$D7#$D8#$D9#117#$85#$DA#$DB#$DC#$DD,
    {cpMac} #135#231#138#128#139#137#147#145#142#131#158#157#146#234#190#189#188#187#203#197#151#238#153#239#154#133#218#217#222#219#228#225#233#232#156#242#243#241#159#134#249#248#236#235,
    {cpDOS} #160#065#132#142#099#067#100#068#130#144#101#069#161#073#108#076#108#076#110#078#162#079#147#079#148#153#114#082#114#082#115#083#116#084#163#085#117#085#129#154#121#089#122#090,
    {cpIBM2} #160#181#132#142#159#171#212#210#130#144#216#183#161#214#146#145#150#149#229#213#162#224#147#226#148#153#234#232#254#253#231#230#156#155#163#233#133#222#129#154#236#237#166#165,
    {cp850} #160#181#132#142#$63#$43#$64#$44#$82#$90#$65#$45#$A1#$D6#$6C#$4C#$6C#$4C#$6E#$4E#$A2#$E0#$93#$E2#$94#$99#$72#$52#$72#$52#$73#$53#$74#$54#$A3#$E9#$75#$55#$81#$9A#$EC#$ED#$7A#$5A,
    {cpAscii} #097#065#097#065#099#067#100#068#101#069#101#069#105#073#108#076#108#076#110#078#111#079#111#079#111#079#114#082#114#082#115#083#116#084#117#085#117#085#117#085#121#089#122#090,
    {cpAny} #097#065#097#065#099#067#100#068#101#069#101#069#105#073#108#076#108#076#110#078#111#079#111#079#111#079#114#082#114#082#115#083#116#084#117#085#117#085#117#085#121#089#122#090);
   {TODO add weights}
{Order:
   žؚݞ
  detectWeight: array[1..MaxChars] of integer = (
  }
   // Specials conversion table
  MaxSpecials = 15;
  cvtSpecials:array[TCodePage] of array[1..MaxSpecials] of Char = (
    {cp895} #253#252#249#250#034#248#$FF#$C4#$C4#$60#$27#$2C#$22#$22#$22,
    {cp1250} #050#110#183#183#168#176#$A0#$96#$97#$91#$92#$82#$93#$94#$84,
    {cp852} #050#110#250#250#249#248#$FF#$C4#$C4#$60#$27#$2C#$22#$22#$22,
    {cpISO88592} #050#110#183#183#168#176#$A0#$2D#$2D#$60#$27#$2C#$22#$22#$22,
    {cpISO88591} #050#110#183#183#168#176#$A0#$2D#$2D#$60#$27#$2C#$22#$22#$22,
    {cpKOI8} #050#110#046#046#168#176#$A0#$B6#$B6#$60#$27#$2C#$22#$22#$22,
    {cpCork} #050#110#046#046#034#111#$20#$0F#$16#$60#$27#$2C#$10#$11#$12,
    {cpDragon} #050#110#046#046#034#177#$20#$2D#$2D#$60#$27#$2C#$22#$22#$22,
    {cpMac} #050#110#046#046#179#111#$CA#$D0#$D1#$D4#$D5#$E2#$D2#$D2#$E3,
    {cpDOS} #253#252#249#250#034#248#$FF#$C4#$C4#$60#$27#$2C#$22#$22#$22,
    {cpIBM2} #050#110#250#250#249#248#$FF#$C4#$C4#$60#$27#$2C#$22#$22#$22,
    {cp850} #253#252#249#250#034#248#$FF#$C4#$C4#$60#$27#$2C#$22#$22#$22,
    {cpAscii} #050#110#046#046#034#111#$20#$2D#$2D#$60#$27#$2C#$22#$22#$22,
    {cpAny} #050#110#046#046#034#111#$20#$2D#$2D#$60#$27#$2C#$22#$22#$22);

   // Borders conversion table
  MaxBorders = 49;
  cvtBorders:array[TCodePage] of array[1..MaxBorders] of Char = (
    {cp895} #196#179#205#186#219#178#177#176#220#221#222#223#254#218#194#191#195#197#180#192#193#217#201#203#187#204#206#185#200#202#188#214#210#183#199#215#182#211#208#189#213#209#184#198#216#181#212#207#190,
    {cp1250} #045#124#045#124#035#035#042#042#042#042#042#042#149#043#043#043#043#043#043#043#043#043#043#043#043#043#043#043#043#043#043#043#043#043#043#043#043#043#043#043#043#043#043#043#043#043#043#043#043,
    {cp852} #196#179#205#186#219#178#177#176#220#219#219#223#245#218#194#191#195#197#180#192#193#217#201#203#187#204#206#185#200#202#188#201#196#187#186#186#186#200#196#188#201#205#187#179#179#179#200#205#188,
    {cpISO88592} #045#124#045#124#035#035#042#042#042#042#042#042#149#043#043#043#043#043#043#043#043#043#043#043#043#043#043#043#043#043#043#043#043#043#043#043#043#043#043#043#043#043#043#043#043#043#043#043#043,
    {cpISO88591} #045#124#045#124#035#035#042#042#042#042#042#042#149#043#043#043#043#043#043#043#043#043#043#043#043#043#043#043#043#043#043#043#043#043#043#043#043#043#043#043#043#043#043#043#043#043#043#043#043,
    {cpKOI8} #045#124#045#124#035#035#042#042#042#042#042#042#042#043#043#043#043#043#043#043#043#043#043#043#043#043#043#043#043#043#043#043#043#043#043#043#043#043#043#043#043#043#043#043#043#043#043#043#043,
    {cpCork} #045#124#045#124#035#035#042#042#042#042#042#042#149#043#043#043#043#043#043#043#043#043#043#043#043#043#043#043#043#043#043#043#043#043#043#043#043#043#043#043#043#043#043#043#043#043#043#043#043,
    {cpDragon} #045#124#045#124#035#035#042#042#042#042#042#042#042#043#043#043#043#043#043#043#043#043#043#043#043#043#043#043#043#043#043#043#043#043#043#043#043#043#043#043#043#043#043#043#043#043#043#043#043,
    {cpMac} #045#124#045#124#035#035#042#042#042#042#042#042#042#043#043#043#043#043#043#043#043#043#043#043#043#043#043#043#043#043#043#043#043#043#043#043#043#043#043#043#043#043#043#043#043#043#043#043#043,
    {cpDOS} #196#179#205#186#219#178#177#176#220#219#219#223#254#218#194#191#195#197#180#192#193#217#201#203#187#204#206#185#200#202#188#201#196#187#186#186#186#200#196#188#201#205#187#179#179#179#200#205#188,
    {cpIBM2} #196#179#205#186#219#178#177#176#220#219#219#223#254#218#194#191#195#197#180#192#193#217#201#203#187#204#206#185#200#202#188#201#196#187#186#186#186#200#196#188#201#205#187#179#179#179#200#205#188,
    {cp850} #196#179#205#186#219#178#177#176#220#219#219#223#254#218#194#191#195#197#180#192#193#217#201#203#187#204#206#185#200#202#188#201#196#187#186#186#186#200#196#188#201#205#187#179#179#179#200#205#188,
    {cpAscii} #045#124#045#124#035#035#042#042#042#042#042#042#042#043#043#043#043#043#043#043#043#043#043#043#043#043#043#043#043#043#043#043#043#043#043#043#043#043#043#043#043#043#043#043#043#043#043#043#043,
    {cpAny} #045#124#045#124#035#035#042#042#042#042#042#042#042#043#043#043#043#043#043#043#043#043#043#043#043#043#043#043#043#043#043#043#043#043#043#043#043#043#043#043#043#043#043#043#043#043#043#043#043);
var
  cvtTablePointers:array[TCodePage] of Pointer;
  cvtSpecialsPointers:array[TCodePage] of Pointer;
  cvtBordersPointers:array[TCodePage] of Pointer;

const
   // Following chars are used in DOS to make borders we should ignore this
  {Border order:
    1--2--3
    |  |  |
    4--5--6
    |  |  |
    7--8--9

    10 -
    11 |
  }
  BorderDataSize = 11;
  SingleBorders:array[1..BorderDataSize] of Char = '¿Ŵĳ'; {|-}
  DoubleBorders:array[1..BorderDataSize] of Char = '˻ιʼͺ'; {||=}
  Mixed1Borders:array[1..BorderDataSize] of Char = 'ҷ׶нĺ'; {||-}
  Mixed2Borders:array[1..BorderDataSize] of Char = 'ѸصϾͳ'; {|=}
  {Boxes order:
  1 light
  2 darker
  3 dark
  4 full
  5 bottom half full
  6 left half full
  7 right half full
  8 upper half full
  9,10,11 stuff to have same size as other border arrays
  }
  Boxes:array[1..BorderDataSize] of Char = '';

  WhiteSpaces:array[1..BorderDataSize] of Char = #32#13#10#9#0#0#0#0#0#0#0;

  EOLns:array[EOLMakeCR..EOLMakeCRLF] of string[2] = (#13, #10, #13#10);
  EOLnsLen:array[EOLMakeCR..EOLMakeCRLF] of Integer = (1, 1, 2);
  CRLFs:array[TCodePage] of TEOL = (
    EOLMakeCRLF,
    EOLMakeCRLF,
    EOLMakeCRLF,
    EOLMakeLF,
    EOLMakeLF,
    EOLMakeCRLF,
    EOLMakeLF,
    EOLMakeCRLF,
    EOLMakeCR,
    EOLMakeCRLF,
    EOLMakeCRLF,

    EOLMakeCRLF,

    EOLMakeCRLF,
    EOLMakeCRLF);


implementation
uses Windows,
  SysUtils,
  TBfThStrm;
const
// Following constants defined because of FPC compatibility
  fmShareExclusive = $0010;
  fmShareDenyWrite = $0020;

function LoCase (Ch:Char):Char; assembler;
asm
   CMP   AL,'A'
   JB    @end
   CMP   AL,'Z'
   JA    @end
   ADD   AL,'a'-'A'
@end:
end;

function FastPos (P:Pointer; BufLen:Cardinal; Chr:Char):Cardinal; assembler;
// Searches for char Chr in buffer at P with length BufLen
// Returns: 0 - not found else position of Chr
asm
        PUSH    EDI
        PUSH    EAX          //P is in EAX
        POP     EDI
        MOV     AL,CL        //in CL is Chr
        MOV     ECX,EDX      //in EDX is BufLen
        REPNE   SCASB
        MOV     EAX,0
        JNE     @@1
        MOV     EAX,EDX      //in EDX is BufLen
        SUB     EAX,ECX      //in ECX if position where char was found
@@1:    POP     EDI
end;

procedure GenerateConvertArray (const CP1, CP2:TCodePage; const C:TCase; var CA:TConvertArray);
var
  I:Integer;
begin
  if C = CaseNoChange then
    for I := 0 to 255 do
      CA[Char (I)] := Char (I)
  else
    if C = CaseLow then
      for I := 0 to 255 do
        CA[Char (I)] := LoCase (Char (I))
    else
      for I := 0 to 255 do
        CA[Char (I)] := UpCase (Char (I));
  if C = CaseNoChange then
  begin
    for I := 1 to MaxChars do
      if cvtTable[CP1][I] > #127 then
        CA[cvtTable[CP1][I]] := cvtTable[CP2][I]
  end
  else
    if C = CaseLow then
    begin
      for I := 1 to (MaxChars div 2) do
      begin
        if cvtTable[CP1][2 * I] > #127 then
          CA[cvtTable[CP1][2 * I]] := cvtTable[CP2][(2 * I) - 1];
        if cvtTable[CP1][(2 * I) - 1] > #127 then
          CA[cvtTable[CP1][(2 * I) - 1]] := cvtTable[CP2][(2 * I) - 1];
      end
    end
    else
    begin
      for I := 1 to (MaxChars div 2) do
      begin
        if cvtTable[CP1][2 * I] > #127 then
          CA[cvtTable[CP1][2 * I]] := cvtTable[CP2][2 * I];
        if cvtTable[CP1][(2 * I) - 1] > #127 then
          CA[cvtTable[CP1][(2 * I) - 1]] := cvtTable[CP2][2 * I];
      end;
    end;
  if ConvertSpecials then
    for I := 1 to MaxSpecials do
      if cvtSpecials[CP1][I] > #127 then
        CA[cvtSpecials[CP1][I]] := cvtSpecials[CP2][I];
  if ConvertSpecials then
    for I := 1 to MaxBorders do
      if cvtBorders[CP1][I] > #127 then
        CA[cvtBorders[CP1][I]] := cvtBorders[CP2][I];
end;

function ConvertChar (const Ch:Char; const CP1, CP2:TCodePage; const ChangeCase:TCase):Char;
var
  I:Cardinal;
begin
  if ChangeCase = CaseNoChange then
  begin
    Result := Ch;
    if Ch < #128 then Exit;
    I := FastPos (cvtTablePointers[CP1], MaxChars, Ch);
    if I <> 0 then
    begin
      Result := cvtTable[CP2][I];
      Exit;
    end;
    if ConvertBorders then
    begin
      I := FastPos (cvtBordersPointers[CP1], MaxBorders, Ch);
      if I <> 0 then
      begin
        Result := cvtBorders[CP2][I];
        Exit;
      end;
    end;
    if ConvertSpecials then
    begin
      I := FastPos (cvtSpecialsPointers[CP1], MaxSpecials, Ch);
      if I <> 0 then
      begin
        Result := cvtSpecials[CP2][I];
      end;
    end;
  end

  else
    if ChangeCase = CaseLow then
    begin
      Result := LoCase (Ch);
      if Ch < #128 then Exit;
      I := FastPos (cvtTablePointers[CP1], MaxChars, Ch);
      if I <> 0 then
      begin
        if UpCase (cvtTable[cpAscii][I]) = cvtTable[cpAscii][I] then Dec (I);
        Result := cvtTable[CP2][I];
        Exit;
      end;
      if ConvertBorders then
      begin
        I := FastPos (cvtBordersPointers[CP1], MaxBorders, Ch);
        if I <> 0 then
        begin
          Result := cvtBorders[CP2][I];
          Exit;
        end;
      end;
      if ConvertSpecials then
      begin
        I := FastPos (cvtSpecialsPointers[CP1], MaxSpecials, Ch);
        if I <> 0 then
        begin
          Result := cvtSpecials[CP2][I];
        end;
      end;
    end

    else
    begin
      Result := UpCase (Ch);
      if Ch < #128 then Exit;
      I := FastPos (cvtTablePointers[CP1], MaxChars, Ch);
      if I <> 0 then
      begin
        if LoCase (cvtTable[cpAscii][I]) = cvtTable[cpAscii][I] then Inc (I);
        Result := cvtTable[CP2][I];
        Exit;
      end;
      if ConvertBorders then
      begin
        I := FastPos (cvtBordersPointers[CP1], MaxBorders, Ch);
        if I <> 0 then
        begin
          Result := cvtBorders[CP2][I];
          Exit;
        end;
      end;
      if ConvertSpecials then
      begin
        I := FastPos (cvtSpecialsPointers[CP1], MaxSpecials, Ch);
        if I <> 0 then
        begin
          Result := cvtSpecials[CP2][I];
        end;
      end;
    end;
end; {ConvertChar}


function ConvertCharNoCase (const Ch:Char; const CP1, CP2:TCodePage):Char;
var
  I:Cardinal;
begin
  Result := Ch;
  if Ch < #128 then Exit;
  I := FastPos (cvtTablePointers[CP1], MaxChars, Ch);
  if I <> 0 then
  begin
    Result := cvtTable[CP2][I];
    Exit;
  end;
  if ConvertBorders then
  begin
    I := FastPos (cvtBordersPointers[CP1], MaxBorders, Ch);
    if I <> 0 then
    begin
      Result := cvtBorders[CP2][I];
      Exit;
    end;
  end;
  if ConvertSpecials then
  begin
    I := FastPos (cvtSpecialsPointers[CP1], MaxSpecials, Ch);
    if I <> 0 then
    begin
      Result := cvtSpecials[CP2][I];
//      Exit;
    end;
  end;
end; {ConvertCharNoCase}

function DetectCodepage (const FileName:string):TCodePage;
var
  Buffer:PChar;
  F:file of Char;
  fs, RS, I, g, gp, gn:Cardinal;
  Counts:array[#128..#255] of Integer;
  CodeCounts:array[TCodePage] of Integer;
  Ch, ChP, ChN:Char;
  CP:TCodePage;
  WasBorder:Boolean;
begin
  Result := cpAscii;
  WasBorder := False;
  if not (FileExists (FileName)) then
    raise Exception.CreateFmt ('CodeConvTools: file error'#13'File %s doesn''t exist!',
      [FileName]);
  for CP := FirstCodePage to LastCodePage do
    CodeCounts[CP] := 0;
  for Ch := #128 to #255 do
    Counts[Ch] := 0;

//Fill buffer with first DetectRead chars of file
  AssignFile (F, FileName);
  FileMode := 0; //Read only
  Reset (F);
  fs := FileSize (F);
  if fs = 0 then begin
    Result := cpAscii;
    Close (F);
    Exit;
  end;

  if fs > DetectRead then fs := DetectRead;
  try
    GetMem (Buffer, fs);
  except
    raise
      Exception.Create ('CodeConvTools: memory error'#13'Not enough memory to detect file code page!');
  end;
  try
    BlockRead (F, Buffer^, fs, RS);
  finally
    Close (F);
  end;

// First char we must do alone - we mustn't, one char we can forget ;-)
  Ch := Buffer[0];
  ChN := Buffer[1];
// Detect chars
  for I := 2 to RS + 1 do
  begin
    ChP := Ch;
    Ch := ChN;
    ChN := Buffer[I];
    if (Ch > #127) and ((Ch <> ChN) or (Ch <> ChP)) then
    begin
      if not (IgnoreBorders) then
        Inc (Counts[Ch])
      else
      begin
        g := FastPos (@SingleBorders, 5 * BorderDataSize, Ch);
        gn := FastPos (@SingleBorders, 6 * BorderDataSize, ChN);
        gp := FastPos (@SingleBorders, 6 * BorderDataSize, ChP);
        if (g * gn * gp) <> 0 then
        begin
          g := g mod BorderDataSize;
          gn := gn mod BorderDataSize;
          gp := gp mod BorderDataSize;
          if
            ((g = 5) and (gn = 5)) or
            ((g = gp) and (gn = 5)) or
            ((g = 5) and (g = gn)) or
            ((g = gp) and (g = gn))
            then
            WasBorder := True
          else
            Inc (Counts[Ch]);
        end
        else
          Inc (Counts[Ch]);
      end
    end;
  end;
  FreeMem (Buffer, fs);

// Detect CodePage
  for CP := FirstCodePage to Pred (LastCodePage) do
    for Ch := #128 to #255 do
    begin
      I := FastPos (cvtTablePointers[CP], MaxChars, Ch);
      if I > 0 then
      begin
        Inc (CodeCounts[CP], Counts[Ch]);
        {TODO wieght sum (may be)}
      end;
    end;

  for CP := FirstCodePage to Pred (LastCodePage) do
    if CodeCounts[CP] > CodeCounts[cpAscii] then
    begin
      CodeCounts[cpAscii] := CodeCounts[CP];
      Result := CP;
    end;
  if (Result = cpAscii) and (WasBorder) then Result := cpDOS;
end; {DetectCodePage}



function ConvertString (const S:string; const CP1, CP2:TCodePage; const ChangeCase:TCase):
string;
var
  I:Integer;
begin
  Result := S;
  for I := 1 to Length (S) do
    Result[I] := ConvertChar (S[I], CP1, CP2, ChangeCase);
end; {ConvertString}



function ConvertFile (FileIn, FileOut:string; var CP1:TCodePage; const CP2:TCodePage; const ChangeCase:TCase; ChangeEOL:TEOL):Cardinal;
var
  LastDone, DoneStep:Cardinal;
  ConvertThread:TConvertThread;

begin
  ConvertThread := TConvertThread.Create (nil, nil, FileIn, FileOut, CP1, CP2, ChangeCase, ChangeEOL);
  ConvertThread.Resume;
  LastDone := 0;
  while ((not (ConvertThread.WasTerminated)) and ((ConvertThread.InputFileSize = 0))) do
    Sleep (10);
  if ProgressStep = 0 then
    DoneStep := ConvertThread.InputFileSize div 100
  else
    DoneStep := ProgressStep;
  while not (ConvertThread.WasTerminated) do
    if ((LastDone + DoneStep) < ConvertThread.CurrentPosition) then
    begin
      LastDone := ConvertThread.CurrentPosition;
      if Assigned (ProgresProc) then ProgresProc (LastDone, ConvertThread.InputFileSize);
      if Assigned (ProgresProc_) then ProgresProc_ (LastDone, ConvertThread.InputFileSize);
    end
    else
      Sleep (100);
  Result := ConvertThread.CurrentPosition;
  if Assigned (ProgresProc) then ProgresProc (ConvertThread.InputFileSize, ConvertThread.InputFileSize);
  if Assigned (ProgresProc_) then ProgresProc_ (ConvertThread.InputFileSize, ConvertThread.InputFileSize);
  ConvertThread.Free;
end;



{
********************************************************************************
********************************************************************************
********************************************************************************
********************************************************************************
TConvertThread implementation begins
********************************************************************************
********************************************************************************
********************************************************************************
********************************************************************************
}


procedure TConvertThread.CallProgress;
begin
  ThreadProgProc (CurrentPosition, InputFileSize);
end;

procedure TConvertThread.CallBeforeTerminateProc;
begin
  ThreadBeforeTerminateProc;
end;

procedure TConvertThread.Execute;
var
  FIn, FOut:THandle;
  ReadThread:TBufferedHandleReadThread;
  WriteThread:TBufferedHandleWriteThread;

  Delta:Cardinal;
  IgnoreLF:Boolean;
  Chr:Char;

  CurrentBuffer:Integer;
  CurrentWriteBuffer:Integer;
  WriteBuffer1Pos, WriteBuffer2Pos:Cardinal;

  Written:Boolean;

  ConvertArray:TConvertArray;

  procedure PutCharToBuffer (const Chr:Char);
  begin
    Written := False;
    repeat
      if (CurrentWriteBuffer = 1) and (WriteThread.Buffer1End = 0) then
      begin
        WriteThread.WriteBuffer1[WriteBuffer1Pos] := Chr;
        Inc (WriteBuffer1Pos);
        Written := True;
        if WriteBuffer1Pos > SizeOf (WriteThread.WriteBuffer1) then
        begin
          WriteThread.Buffer1End := SizeOf (WriteThread.WriteBuffer1);
          WriteBuffer1Pos := 1;
          CurrentWriteBuffer := 2;
        end;
      end
      else
        if (CurrentWriteBuffer = 2) and (WriteThread.Buffer2End = 0) then
        begin
          WriteThread.WriteBuffer2[WriteBuffer2Pos] := Chr;
          Inc (WriteBuffer2Pos);
          Written := True;
          if WriteBuffer2Pos > SizeOf (WriteThread.WriteBuffer2) then
          begin
            WriteThread.Buffer2End := SizeOf (WriteThread.WriteBuffer2);
            WriteBuffer2Pos := 1;
            CurrentWriteBuffer := 1;
          end;
        end;
    until (Written);

  end;

var F:file of Char;
  fs:Integer;

begin
  WasTerminated := False;
  GenerateConvertArray (FCp1, FCp2, FChangeCase, ConvertArray);
  CurrentPosition := 0;
  CurrentBuffer := 1;
  CurrentWriteBuffer := 1;
  WriteBuffer1Pos := 1;
  WriteBuffer2Pos := 1;
  IgnoreLF := False;
  Delta := 0;

  CompareFiles := AnsiCompareText (FFileIn, FFileOut) = 0;

  if FChangeEOL = EOLMakeAsCP then FChangeEOL := CRLFs[FCp2];
  try
    if FCp2 = cpAny then
      raise
        Exception.Create ('CodeConvTools: ConvertFile parameters error'#13'Can''t convert to any code page.'#13'You must specify exact target code page!');
    if FCp1 = cpAny then FCp1 := DetectCodepage (FFileIn);
    FFileIn := ExpandFileName (FFileIn);
    FFileOut := ExpandFileName (FFileOut);

    if not (FileExists (FFileIn)) then
      raise Exception.CreateFmt ('CodeConvTools: file error'#13'File %s doesn''t exist!', [FFileIn]);

    AssignFile (F, FFileIn);
    FileMode := 0; //Read only
    Reset (F);
    fs := FileSize (F);
    Close (F);

    if fs = 0 then begin
      if CompareFiles then Exit;
      FOut := CreateFile (PChar (OutName), GENERIC_WRITE, 0, nil, CREATE_ALWAYS, 0, 0);
      if FOut = INVALID_HANDLE_VALUE then
      begin
        raise
          Exception.CreateFmt ('CodeConvTools: file error'#13'Error while opening file %s for writing!', [OutName]);
      end;
      CloseHandle (FOut);
    end;



    FIn := CreateFile (PChar (FFileIn), GENERIC_READ, FILE_SHARE_READ, nil, OPEN_EXISTING, FILE_FLAG_SEQUENTIAL_SCAN, 0);
    if FIn = INVALID_HANDLE_VALUE then
      raise
        Exception.CreateFmt ('CodeConvTools: file error'#13'Error while opening file %s for reading!', [FFileIn]);

    InputFileSize := GetFileSize (FIn, nil);
    ThreadProgressStep := InputFileSize div 100;

    if CompareFiles and ((FCp2 = FCp1) or (FCp1 = cpAscii)) and (FChangeCase = CaseNoChange) and (FChangeEOL = EOLNoChange) then
    begin
      CurrentPosition := InputFileSize;
      CloseHandle (FIn);
      Exit; {Nothing to do here...}
    end;

    if CompareFiles then
    begin
      OutName := ExtractFilePath (FFileOut) + TemporaryFileName;
      FOut := CreateFile (PChar (OutName), GENERIC_WRITE, 0, nil, CREATE_ALWAYS, FILE_ATTRIBUTE_TEMPORARY, 0);
      if FOut = INVALID_HANDLE_VALUE then
      begin
        CloseHandle (FIn);
        raise
          Exception.CreateFmt ('CodeConvTools: file error'#13'Error while opening file %s for writing!', [OutName]);
      end;
    end
    else
    begin
      OutName := FFileOut;
      FOut := CreateFile (PChar (OutName), GENERIC_WRITE, 0, nil, CREATE_ALWAYS, 0, 0);
      if FOut = INVALID_HANDLE_VALUE then
      begin
        CloseHandle (FIn);
        raise
          Exception.CreateFmt ('CodeConvTools: file error'#13'Error while opening file %s for writing!', [OutName]);
      end;
    end;

    ReadThread := TBufferedHandleReadThread.Create (FIn);
    ReadThread.Resume;
    WriteThread := TBufferedHandleWriteThread.Create (FOut);
    WriteThread.Resume;

    if Assigned (ThreadProgProc) then Synchronize (CallProgress);

    try
{vvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvv(CONVERT LOOP BEGIN)vvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvv}
      while (CurrentPosition < InputFileSize) do
      begin
        if Terminated then Exit;
        if {read buffers are empty}
          ((CurrentBuffer = 2) and (ReadThread.Buffer2Pos > ReadThread.Buffer2End)) or
          ((CurrentBuffer = 1) and (ReadThread.Buffer1Pos > ReadThread.Buffer1End)) then
        begin
          Sleep (1);
          Continue;
        end;

        if (CurrentBuffer = 1) then
        begin
            // get char from buffer 1
          Chr := ReadThread.ReadBuffer1[ReadThread.Buffer1Pos];
          Inc (CurrentPosition);
          Inc (ReadThread.Buffer1Pos);
          if ReadThread.Buffer1Pos > ReadThread.Buffer1End then CurrentBuffer := 2;
        end
        else
        begin
            // get char from buffer 2
          Chr := ReadThread.ReadBuffer2[ReadThread.Buffer2Pos];
          Inc (CurrentPosition);
          Inc (ReadThread.Buffer2Pos);
          if ReadThread.Buffer2Pos > ReadThread.Buffer2End then CurrentBuffer := 1;
        end;

        //convert and write char
        if (FChangeEOL = EOLNoChange) or ((Chr <> #10) and (Chr <> #13)) then
        begin
          IgnoreLF := False;
          PutCharToBuffer (ConvertArray[Chr]);
        end
        else
          if (Chr = #13) then
          begin
            IgnoreLF := True;
            if EOLnsLen[FChangeEOL] = 1 then
              PutCharToBuffer (EOLns[FChangeEOL][1])
            else
            begin
              PutCharToBuffer (EOLns[FChangeEOL][1]);
              PutCharToBuffer (EOLns[FChangeEOL][2]);
            end;
          end
          else
          begin
            if not (IgnoreLF) then
            begin
              if EOLnsLen[FChangeEOL] = 1 then
                PutCharToBuffer (EOLns[FChangeEOL][1])
              else
              begin
                PutCharToBuffer (EOLns[FChangeEOL][1]);
                PutCharToBuffer (EOLns[FChangeEOL][2]);
              end;
            end
            else
              IgnoreLF := False;
          end;

        Inc (Delta);
        if (Delta > ThreadProgressStep) then
        begin
          if Assigned (ThreadProgProc) then Synchronize (CallProgress);
          Delta := 0;
        end;
      end; {while}
{^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^(CONVERT LOOP END)^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^}
      if Assigned (ThreadProgProc) then Synchronize (CallProgress);

    finally
      try
        // Try to write data if buffer!
        if (CurrentWriteBuffer = 1) and (WriteBuffer1Pos > 1) then
          WriteThread.Buffer1End := WriteBuffer1Pos - 1
        else
          if (CurrentWriteBuffer = 2) and (WriteBuffer2Pos > 1) then
            WriteThread.Buffer2End := WriteBuffer2Pos - 1;
      finally
        try
          ReadThread.Terminate;
          ReadThread.WaitFor;
          ReadThread.Free;
          WriteThread.Terminate;
          WriteThread.WaitFor;
          WriteThread.Free;
        finally
          CloseHandle (FIn);
          CloseHandle (FOut);
          if CompareFiles then
          begin
            //Handle read only files:
            FileSetAttr (OutName, FileGetAttr (FFileIn));
            FileSetAttr(FFileIn, 0);
            
            DeleteFile (FFileIn);
            OutName := OutName + #0;
            FFileOut := FFileOut + #0;
            MoveFile (@OutName[1], @FFileOut[1]);
          end;
        end;
      end;
    end;

  finally
    if Assigned (ThreadBeforeTerminateProc) then Synchronize (CallBeforeTerminateProc);
    WasTerminated := True;
  end;
end;



constructor TConvertThread.Create;
begin
  ThreadProgProc := ProgProc;
  BeforeTerminateProc := BeforeTerminateProc;
  FFileIn := FileIn;
  FFileOut := FileOut;
  FCp1 := CP1;
  FCp2 := CP2;
  FChangeCase := ChangeCase;
  FChangeEOL := ChangeEOL;
  inherited Create (True);
  FreeOnTerminate := False;
end;

procedure Init;
var
  CP:TCodePage;
begin
  for CP := FirstCodePage to LastCodePage do
  begin
    cvtTablePointers[CP] := @cvtTable[CP][1];
    cvtSpecialsPointers[CP] := @cvtSpecials[CP][1];
    cvtBordersPointers[CP] := @cvtSpecials[CP][1];
  end;
end;

initialization
  Init;
end.

